#include "common.h"
#include "bp35a1-client.h"
#include "edispz-controller.h"

enum
{
  BP35A1_CLIENT_PROP_0,
  BP35A1_CLIENT_PROP_DEVICE,
  BP35A1_CLIENT_PROP_READ_BUF_SIZE
};

G_DEFINE_TYPE (Bp35a1Client, bp35a1_client, G_TYPE_OBJECT)

static void bp35a1_register_free (gpointer reg);
static int bp35a1_client_restore_serial (Bp35a1Client * client);
static int bp35a1_client_setup_serial (Bp35a1Client * client);
static gboolean bp35a1_client_write_register (Bp35a1Client * client, gchar * reg_name, gchar * value);
static gchar * bp35a1_client_read_register (Bp35a1Client * client, gchar * reg_name);
static Bp35a1Message ** bp35a1_client_receive_messages_timeout (Bp35a1Client * client, struct timeval *timeout);
static gint bp35a1_client_whitelist_compare (gconstpointer a, gconstpointer b);

static gchar *register_list[] = {
  "S02", "S03", "S07", "S0A", "S0B", "S15", "S16", "S17", "S1C",
  "SA1", "SA2", "SA9", "SF0", "SFB", "SFD", "SFE", "SFF", NULL
};

static void
bp35a1_client_get_property (GObject    *object,
                            guint       property_id,
                            GValue     *value,
                            GParamSpec *pspec)
{
  g_return_if_fail (BP35A1_IS_CLIENT(object));

  switch (property_id)
    {
    case BP35A1_CLIENT_PROP_DEVICE:
      g_value_set_string (value, bp35a1_client_get_device_name((Bp35a1Client *)object));
      break;
    case BP35A1_CLIENT_PROP_READ_BUF_SIZE:
      g_value_set_uint (value, bp35a1_client_get_read_buf_size ((Bp35a1Client *)object));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
bp35a1_client_set_property (GObject      *object,
                            guint         property_id,
                            const GValue *value,
                            GParamSpec   *pspec)
{
  g_return_if_fail (BP35A1_IS_CLIENT(object));

  switch (property_id)
    {
    case BP35A1_CLIENT_PROP_DEVICE:
      bp35a1_client_set_device_name((Bp35a1Client *)object, g_value_get_string (value));
      break;
    case BP35A1_CLIENT_PROP_READ_BUF_SIZE:
      bp35a1_client_set_read_buf_size((Bp35a1Client *)object, g_value_get_uint (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
bp35a1_client_finalize (GObject *object)
{
  if (G_OBJECT_CLASS (bp35a1_client_parent_class))
    G_OBJECT_CLASS (bp35a1_client_parent_class)->finalize (object);
}

static void
bp35a1_client_class_init (Bp35a1ClientClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->get_property = bp35a1_client_get_property;
  gobject_class->set_property = bp35a1_client_set_property;
  gobject_class->finalize = bp35a1_client_finalize;

  g_object_class_install_property (gobject_class, BP35A1_CLIENT_PROP_DEVICE,
      g_param_spec_string ("device", "Device", "BP35A1 device file",
          BP35A1_CLIENT_DEVICE, G_PARAM_READWRITE | G_PARAM_CONSTRUCT_ONLY | G_PARAM_STATIC_STRINGS));

  g_object_class_install_property (gobject_class, BP35A1_CLIENT_PROP_READ_BUF_SIZE,
      g_param_spec_uint ("read-buffer-size", "Read buffer size", "Read buffer size",
          1, G_MAXUINT, BP35A1_CLIENT_READ_BUF_SIZE, G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS ));
}

static void
bp35a1_client_init (Bp35a1Client *self)
{
  self->fd = -1;
  self->read_buf_size = 0;
  self->read_timeout.tv_sec = BP35A1_CLIENT_READ_TIMEOUT_SEC;
  self->read_timeout.tv_usec = BP35A1_CLIENT_READ_TIMEOUT_USEC;
  self->register_table = NULL;
  self->whitelist = NULL;
  self->encrypt_option[0] =  '0';
  self->encrypt_option[1] =  '\0';
}

Bp35a1Client *
bp35a1_client_new (void)
{
  return g_object_new (BP35A1_TYPE_CLIENT, NULL);
}

gchar *
bp35a1_client_get_device_name (Bp35a1Client *client)
{
  gchar *result;

  g_return_val_if_fail (BP35A1_IS_CLIENT (client), NULL);

  result = client->device_name;

  return result;
}

void
bp35a1_client_set_device_name (Bp35a1Client *client, const gchar *string)
{
  g_return_if_fail (BP35A1_IS_CLIENT (client));
  g_return_if_fail (string != NULL);

  if (client->device_name != NULL) {
    if (strcmp(client->device_name, string) == 0)
      return;

    g_free (client->device_name);
  }

  client->device_name = g_strdup (string);
}

guint
bp35a1_client_get_read_buf_size (Bp35a1Client *client)
{
  guint result;

  g_return_val_if_fail (BP35A1_IS_CLIENT (client), 0);

  result = client->read_buf_size;

  return result;
}

void
bp35a1_client_set_read_buf_size (Bp35a1Client *client, guint size)
{
  g_return_if_fail (BP35A1_IS_CLIENT (client));
  g_return_if_fail (size > 0);

  if (client->read_buf_size == size)
    return;

  client->read_buf_size = size;
}

gint
bp35a1_client_start (Bp35a1Client *client)
{
  gint i;
  Bp35a1Register *reg;

  g_return_val_if_fail (BP35A1_IS_CLIENT (client), -1);

  client->fd = open (client->device_name, O_RDWR | O_NOCTTY | O_NONBLOCK);
  if (client->fd == -1)
    return -2;

  client->read_timeout.tv_sec = BP35A1_CLIENT_READ_TIMEOUT_SEC;
  client->read_timeout.tv_usec = BP35A1_CLIENT_READ_TIMEOUT_USEC;

  bp35a1_client_setup_serial (client);
  bp35a1_client_write_register (client, "S2",  BP35A1_CLIENT_REG_CH);
  bp35a1_client_write_register (client, "S3",  BP35A1_CLIENT_REG_PAN_ID);
  bp35a1_client_write_register (client, "S0A", BP35A1_CLIENT_REG_HAN_ID);
  bp35a1_client_write_register (client, "S15", BP35A1_CLIENT_REG_BEACON_REQ);

  for (i = 0; register_list[i]; i++) {
    reg = g_malloc (sizeof (Bp35a1Register));
    reg->name = g_strdup (register_list[i]);
    reg->value = bp35a1_client_read_register (client, reg->name);
    client->register_table =
        g_list_append (client->register_table, reg);
  }

  return 0;
}

gint
bp35a1_client_stop (Bp35a1Client *client)
{
  g_return_val_if_fail  (client != NULL, -1);

  if (client->register_table) {
    g_list_free_full (client->register_table, bp35a1_register_free);
  }

  bp35a1_client_restore_serial (client);

  if (client->fd > 0) {
    close (client->fd);
    client->fd = -1;
  }

  if (client->device_name)
    g_free (client->device_name);

  return 0;
}

static void
bp35a1_register_free (gpointer reg)
{
  Bp35a1Register *bp35a1_register = (Bp35a1Register *) reg;

  if (bp35a1_register) {
    if (bp35a1_register->name)
      g_free (bp35a1_register->name);
    if (bp35a1_register->value)
      g_free (bp35a1_register->value);
    g_free (bp35a1_register);
  }
}


static int
bp35a1_client_restore_serial (Bp35a1Client * client)
{
  return tcsetattr (client->fd, TCSANOW, &client->old_tio);
}

static int
bp35a1_client_setup_serial (Bp35a1Client * client)
{
  struct termios tio;
  int ret;

  ret = tcgetattr (client->fd, &client->old_tio);
  if (ret != 0)
    return -1;

  memset (&tio, 0, sizeof (tio));
  tio.c_iflag = IGNBRK | IGNPAR;        // ブレーク文字無視/パリティな
  tio.c_cflag = CS8 | CLOCAL | CREAD;   // フロー制御なし/8bit/非モデム/受信可
  tio.c_cc[VTIME] = 0;          // キャラクタ間タイマー無効
  tio.c_cc[VMIN] = 1;           // 最低1文字送信/受信するまでブロックする
  ret = cfsetspeed (&tio, B115200);       // 入出力ボーレート
  if (ret < 0)
    return -1;

  ret = tcflush (client->fd, TCIFLUSH);
  if (ret < 0)
    return -1;

  ret = tcsetattr (client->fd, TCSANOW, &tio);
  if (ret != 0)
    return -1;

  return 0;
}

static gboolean
bp35a1_client_write_register (Bp35a1Client * client, gchar * reg_name, gchar * value)
{
  Bp35a1Message **msg_array;
  Bp35a1Message *msg;
  Bp35a1MessageType type;
  GString *gstr;
  GByteArray *data;
  struct timeval timeout;
  guint i;
  gboolean ret = FALSE;

  timeout.tv_sec = 1;
  timeout.tv_usec= 0;

  gstr = g_string_new (bp35a1_message_type_get_name (BP35A1_MESSAGE_SKSREG));
  g_string_append_printf (gstr, " %s %s", reg_name, value);
  data = g_byte_array_new_take ((guint8 *)g_strdup (gstr->str), gstr->len);

  msg = bp35a1_message_new ();
  msg = bp35a1_message_overwrite (msg, data);

  g_byte_array_free (data, TRUE);
  g_string_free (gstr, TRUE);

  bp35a1_client_post_message (client, msg);
  msg_array = bp35a1_client_receive_messages_timeout (client, &timeout);

  if (msg_array != NULL) {
    for (i = 0; msg_array[i]; i++) {
      type = bp35a1_message_get_message_type (msg_array[i]);
      if (type == BP35A1_MESSAGE_OK) {
	ret = TRUE;
      }
      g_object_unref (msg_array[i]);
    }
    g_free (msg_array);
  }

  return ret;
}

static gchar *
bp35a1_client_read_register (Bp35a1Client * client, gchar * reg_name)
{
  Bp35a1Message **msg_array;
  Bp35a1Message *msg;
  Bp35a1MessageType type;
  GString *gstr;
  GByteArray *data;
  struct timeval timeout;
  guint i;
  gchar *ret = NULL;

  timeout.tv_sec = 1;
  timeout.tv_usec= 0;

  gstr = g_string_new (bp35a1_message_type_get_name (BP35A1_MESSAGE_SKSREG));
  g_string_append_printf (gstr, " %s", reg_name);
  data = g_byte_array_new_take ((guint8 *)g_strdup (gstr->str), gstr->len);

  msg = bp35a1_message_new ();
  msg = bp35a1_message_overwrite (msg, data);

  g_byte_array_free (data, TRUE);
  g_string_free (gstr, TRUE);

  bp35a1_client_post_message (client, msg);
  msg_array = bp35a1_client_receive_messages_timeout (client, &timeout);

  if (msg_array != NULL) {
    for (i = 0; msg_array[i]; i++) {
      type = bp35a1_message_get_message_type (msg_array[i]);
       if (type == BP35A1_MESSAGE_ESREG && ret == NULL) {
         data = bp35a1_client_get_message_data (client, msg_array[i]);
         ret = (gchar *) g_byte_array_free (data, FALSE);
       }
      g_object_unref (msg_array[i]);
    }
    g_free (msg_array);
  }

  return ret;
}

gint
bp35a1_client_post_message (Bp35a1Client * client, Bp35a1Message * message)
{
  GByteArray *data;
  Bp35a1MessageType type;

  g_return_val_if_fail (client != NULL, -1);
  g_return_val_if_fail (message != NULL, -1);

  data = bp35a1_message_read (message);
  type = bp35a1_message_get_message_type (message);

  xwrite (client->fd, data->data, data->len);
  switch (type) {
    case BP35A1_MESSAGE_SKSENDTO:
      break;
    default:
      xwrite (client->fd, "\r\n", strlen("\r\n"));
      break;
  }

  g_byte_array_free (data, TRUE);
  g_object_unref (message);

  return 0;
}

static Bp35a1Message **
bp35a1_client_receive_messages_timeout (Bp35a1Client * client,
    struct timeval *timeout)
{
  Bp35a1Message **msg_array;
  Bp35a1Message *msg;
  GByteArray *data;
  gpointer buf;
  gsize size;

  int ret;
  fd_set fds;

  FD_ZERO (&fds);
  FD_SET (client->fd, &fds);

  ret = select (client->fd + 1, &fds, NULL, NULL, timeout);

  if (ret == -1) {
    return NULL;
  } else if (ret && FD_ISSET (client->fd, &fds)) {
    buf = g_malloc (BP35A1_CLIENT_READ_BUF_SIZE);
    if (!buf) {
      return NULL;
    }
    size = xread (client->fd, buf, BP35A1_CLIENT_READ_BUF_SIZE);
    if (size == -1) {
      g_free (buf);
      return NULL;
    }

    data = g_byte_array_new_take (buf, size);

    msg = bp35a1_message_new ();
    msg = bp35a1_message_overwrite (msg, data);

    g_byte_array_free (data, TRUE);

    msg_array = bp35a1_message_split (msg);

    g_object_unref(msg);

    return msg_array;
  } else {
    return NULL;
  }
}

Bp35a1Message **
bp35a1_client_receive_messages (Bp35a1Client * client)
{
  struct timeval timeout = client->read_timeout;
  return bp35a1_client_receive_messages_timeout (client, &timeout);
}

GByteArray *
bp35a1_client_get_message_data(Bp35a1Client * client, Bp35a1Message * message)
{
  gboolean p;
  return bp35a1_client_get_message_data_permittable (client, message, &p);
}


GByteArray *
bp35a1_client_get_message_data_permittable (Bp35a1Client * client, Bp35a1Message * message, gboolean *permit)
{
  GByteArray *data = NULL;
  Bp35a1MessageType type;
  *permit = TRUE;

  type = bp35a1_message_get_message_type (message);
  switch (type) {
    case BP35A1_MESSAGE_SKSREG:
    case BP35A1_MESSAGE_SKINFO:
    case BP35A1_MESSAGE_SKSENDTO:
      break;
    case BP35A1_MESSAGE_ERXUDP:
    {
      GByteArray *msg_data;
      gchar **strv;
      gboolean rssi_include = FALSE;
      gint erxudp_header_size = 121;

      msg_data = bp35a1_message_read (message);
      strv = g_strsplit ((gchar *)msg_data->data, " ", -1);
      if (g_strv_length(strv) < 10){
        g_strfreev (strv);
        g_byte_array_free (msg_data, TRUE);
        break;
      }
      if (strcmp (strv[7], "0") == 0 || strcmp (strv[7], "1") == 0)
        rssi_include = TRUE;

      if (client->whitelist != NULL) {
        gchar *ipaddr = strv[1];
        gchar *mac    = strv[5];
        if (!bp35a1_client_whitelist_is_exist (client, ipaddr) &&
	    !bp35a1_client_whitelist_is_exist (client, mac)) {
	  *permit = FALSE;
	  g_debug("e-DISPZ:%s is ignored.", mac);
#ifdef IGNORE_BEFORE_PARSE
          g_byte_array_free (msg_data, TRUE);
          msg_data = NULL;
#endif
        }
      } else {
	*permit = FALSE;
      }

      g_strfreev (strv);

      if (msg_data != NULL) {
        if (msg_data->len > erxudp_header_size) {
          if (!rssi_include)
            g_byte_array_remove_range (msg_data, 0, erxudp_header_size);
          else
            g_byte_array_remove_range (msg_data, 0, erxudp_header_size + 2);

          data = msg_data;
        }
      }
    }
      break;
    case BP35A1_MESSAGE_ESREG:
    {
      GByteArray *msg_data;
      GRegex *reg;
      gchar *str, *s, *r;

      msg_data = bp35a1_message_read (message);
      msg_data = g_byte_array_append (msg_data, (guint8 *)"\0", 1);
      str = (gchar *) g_byte_array_free (msg_data, FALSE);
      s = str + strlen(bp35a1_message_type_get_name(type)) + 1;

      reg = g_regex_new ("\r|\n", 0, 0, NULL);
      r = g_regex_replace (reg, s, -1, 0, "\0", 0, NULL);
      g_regex_unref (reg);

      g_free (str);

      data = g_byte_array_new();
      data = g_byte_array_append (data, (guint8 *) r, strlen(r)+1);

      g_free (r);
    }
      break;
    default:
      break;
  }

  return data;
}

void
bp35a1_client_debug_dump_registers (Bp35a1Client * client)
{
  Bp35a1Register *reg;
  GList *list;

  for (list = client->register_table; list; list = list->next) {
    reg = (Bp35a1Register *)list->data;
    g_debug ("%s: %s", reg->name, reg->value);
  }
}

Bp35a1Message *
bp35a1_client_new_response_message (Bp35a1Client * client,
    Bp35a1Message * src_message, GByteArray * payload)
{
  Bp35a1Message *response;
  GByteArray *msg_data;
  gchar *str, **strv, *ipaddr, *port, *side, *data;

  g_return_val_if_fail (client != NULL, NULL);
  g_return_val_if_fail (src_message != NULL, NULL);
  g_return_val_if_fail (payload != NULL, NULL);

  msg_data = bp35a1_message_read (src_message);
  msg_data = g_byte_array_append (msg_data, (guint8 *) "\0", 1);
  str = (gchar *) g_byte_array_free (msg_data, FALSE);
  strv = g_strsplit (str, " ", -1);
  ipaddr = strv[1];
  port = strv[3];
  side = strv[8];
  if (strcmp (side, "0") != 0 && strcmp (side, "1") != 0)
    side = strv[7];

  data = g_strdup_printf ("%s %d %s %s %s %s %04x ",
      bp35a1_message_type_get_name (BP35A1_MESSAGE_SKSENDTO),
      BP35A1_CLIENT_ECHONET_LITE_HANDLE,
      ipaddr, port, client->encrypt_option, side, payload->len);
  g_byte_array_prepend (payload, (guint8 *) data, strlen (data));
  g_free (data);

  g_strfreev (strv);
  g_free (str);

  response = bp35a1_message_new ();
  response = bp35a1_message_overwrite (response, payload);
  return response;
}

gint
bp35a1_client_whitelist_add (Bp35a1Client * client, const gchar * item)
{
  GString *str = NULL;
  GRegex *reg;
  gchar *s, *p;
  guint i, len;
  gint ret = -1;

  g_return_val_if_fail (item != NULL, -1);

  if (strlen (item) < 12)
    return ret;

  reg = g_regex_new (" |:", 0, 0, NULL);
  s = g_regex_replace (reg, item, -1, 0, "", 0, NULL);
  g_regex_unref (reg);

  len = strlen(s);
  if (len == 32) {         /* IPv6 */
    str = g_string_new (NULL);
    for (i = 0; i < len; i+=4) {
      p = g_strndup (s + i, 4);
      g_string_append_printf (str, "%s:", p);
      g_free (p);
    }
    str = g_string_erase (str, str->len-1, 1);
  } else if (len == 16) {  /* MAC */
    str = g_string_new (s);
  } else if (len == 12) {  /* MAC */
    str = g_string_new (s);
    str = g_string_insert (str, 6, "FFFE");
  }

  if (str != NULL) {
    client->whitelist = g_list_append (client->whitelist, g_string_free (str, FALSE));
    ret = 0;
  }

  g_free (s);

  return ret;
}

gint
bp35a1_client_whitelist_delete (Bp35a1Client * client, const gchar * item)
{
  if (bp35a1_client_whitelist_is_exist(client, item)) {
    client->whitelist = g_list_remove (client->whitelist, item);
    return 0;
  } else {
    return 1;
  }
}

static gint
bp35a1_client_whitelist_compare (gconstpointer a, gconstpointer b)
{
  return strcmp ((gchar *)a, (gchar *)b);
}

gboolean
bp35a1_client_whitelist_is_exist (Bp35a1Client * client, const gchar * item)
{
  GList *target;
  target = g_list_find_custom (client->whitelist, item, bp35a1_client_whitelist_compare);

  if (target)
    return TRUE;
  else
    return FALSE;
}

gint
bp35a1_client_whitelist_get_index(Bp35a1Client * client, const gchar * item)
{
  GList *target;
  target = g_list_find_custom (client->whitelist, item, bp35a1_client_whitelist_compare);
  return g_list_position(client->whitelist, target);
}
