#include "bp35a1-message.h"
#include "bp35a1-raw-message.h"

static Bp35a1MessageTypeHash message_type_table[] = {
  {BP35A1_MESSAGE_SKSREG, "SKSREG"},
  {BP35A1_MESSAGE_SKINFO, "SKINFO"},
  {BP35A1_MESSAGE_SKSENDTO, "SKSENDTO"},
  {BP35A1_MESSAGE_ERXUDP, "ERXUDP"},
  {BP35A1_MESSAGE_ESREG, "ESREG"},
  {BP35A1_MESSAGE_EVENT, "EVENT"},
  {BP35A1_MESSAGE_OK, "OK"},
  {BP35A1_MESSAGE_UNKNOWN, NULL},
};

G_DEFINE_TYPE (Bp35a1Message, bp35a1_message, G_TYPE_OBJECT)

static gint bp35a1_message_addr_compare (gconstpointer a, gconstpointer b);

static void
bp35a1_message_finalize (GObject *object)
{
  Bp35a1Message *msg = BP35A1_MESSAGE(object);

  if (msg->raw != NULL) {
    bp35a1_raw_message_free (msg->raw);
  }

  if (G_OBJECT_CLASS (bp35a1_message_parent_class))
    G_OBJECT_CLASS (bp35a1_message_parent_class)->finalize (object);
}

static void
bp35a1_message_class_init (Bp35a1MessageClass *klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->finalize = bp35a1_message_finalize;
}

static void
bp35a1_message_init (Bp35a1Message *self)
{
  self->raw = NULL;
}

Bp35a1Message *
bp35a1_message_new (void)
{
  return g_object_new (BP35A1_TYPE_MESSAGE, NULL);
}

Bp35a1Message *
bp35a1_message_overwrite (Bp35a1Message * message, GByteArray * data)
{
  g_return_val_if_fail (message != NULL, NULL);

  if (message->raw != NULL)
    bp35a1_raw_message_free (message->raw);

  message->raw = bp35a1_raw_message_new (data->data, data->len);

  return message;
}

GByteArray *
bp35a1_message_read (Bp35a1Message * message)
{
  Bp35a1RawMessage *raw;
  GByteArray *result;

  g_return_val_if_fail (message != NULL, NULL);

  result = g_byte_array_new();

  raw = (Bp35a1RawMessage *)message->raw;
  result = g_byte_array_append (result, raw->data, raw->size);

  return result;
}

gchar *
bp35a1_message_get_mac_address (Bp35a1Message * message)
{
  GByteArray *msg_data;
  gchar **strv;
  gchar *mac = NULL;

  msg_data = bp35a1_message_read (message);
  if (msg_data) {
    strv = g_strsplit ((gchar *)msg_data->data, " ", -1);
    if (g_strv_length(strv) > 5)
      mac = g_strdup(strv[5]);
    g_strfreev(strv);
    g_byte_array_free (msg_data, TRUE);
  }

  return mac;
}


Bp35a1MessageType
bp35a1_message_get_message_type (Bp35a1Message * message)
{
  GRegex *reg;
  gchar *s, *r;
  guint i;
  Bp35a1RawMessage **raw_array;
  Bp35a1MessageType type = BP35A1_MESSAGE_UNKNOWN;

  raw_array = bp35a1_raw_message_split (message->raw, " ", 2);


  r = bp35a1_raw_message_to_string (raw_array[0]);
  reg = g_regex_new ("\r.*|\n.*", 0, 0, NULL);
  s = g_regex_replace (reg, r, -1, 0, "\0", 0, NULL);
  g_regex_unref (reg);
  for (i = 0; message_type_table[i].name; i++)
    if (0 == strcmp (message_type_table[i].name, s))
      break;
  type = message_type_table[i].type;
  g_free (s);
  g_free (r);

  bp35a1_raw_message_freev (raw_array);

  return type;
}

const gchar *bp35a1_message_type_get_name (Bp35a1MessageType type)
{
  guint i;

  for (i = 0; message_type_table[i].name; i++)
    if (message_type_table[i].type == type)
      return message_type_table[i].name;

  return "unknown";
}

static gint
bp35a1_message_addr_compare (gconstpointer a, gconstpointer b)
{
  if (a == b)
    return 0;
  else if (a > b)
    return 1;
  else
    return -1;
}

Bp35a1Message **
bp35a1_message_split (Bp35a1Message * message)
{
  GSList *msg_list = NULL, *mlist;
  Bp35a1Message **msg_array;
  Bp35a1RawMessage *raw;
  Bp35a1Message *msg;
  GByteArray *data;
  gpointer remainder, s;
  guint i, n;

  raw = message->raw;

  remainder = raw->data;
  for (i = 0; message_type_table[i].name; i++) {
    s = (gpointer) strstr (remainder, message_type_table[i].name);
    if (s != NULL) {
      msg_list = g_slist_prepend (msg_list, s);
    }
  }
  msg_list = g_slist_sort (msg_list, bp35a1_message_addr_compare);

  msg_array = g_new (Bp35a1Message *, g_slist_length (msg_list) + 1);

  n = 0;
  for (mlist = msg_list; mlist; mlist = mlist->next) {
    data = g_byte_array_new ();
    msg = bp35a1_message_new();

    if (mlist->next != NULL) {
      data = g_byte_array_append (data, mlist->data, (gsize)(mlist->next->data - mlist->data));
    } else {
      data = g_byte_array_append (data, mlist->data, (gsize)(raw->size - (mlist->data - raw->data)));
    }

    msg_array[n++] = bp35a1_message_overwrite (msg, data);

    g_byte_array_free (data, TRUE);
  }
  msg_array[n] = NULL;

  g_slist_free (msg_list);

  return msg_array;
}
