#include "edispz-controller.h"

static GByteArray *edispz_controller_create_timekeep_packet (EdispzController *
							     controller, gchar *mac, gboolean permit);
static gboolean edispz_controller_is_packet (GByteArray * packet);
static gboolean edispz_controller_is_special_packet (GByteArray * packet);
static void edispz_controller_loop (gpointer * controller);
static gchar *edispz_ack_nack_packet_get_data (GByteArray * packet,
    EdispzPacketDataType data_type);
static gchar *edispz_normal_packet_get_data (guint8 * data,
    EdispzPacketDataType data_type);
static gchar *edispz_special_packet_get_data (guint8 * data,
    EdispzPacketDataType data_type);

static gboolean
is_immediate_mode(guint mode)
{
  if (mode >= 31 && mode <= 39)
    return TRUE;

  return FALSE;
}

static void print_config(EdispzController * controller)
{
  if (is_immediate_mode(controller->operation_mode)) {
    g_info("operation mode: %d(Immediate)", controller->operation_mode);
    g_info("human sensor wait time(second): %d", controller->human_sensor_wait_time);
    g_info("heartbeat time(minute): %d", controller->heartbeat_time);
    g_info("door detect mode: %d", controller->door_detect_mode);
  } else {
    g_info("operation mode: %d(Nomal)", controller->operation_mode);
  }
  g_info("interval time(second): %d", controller->interval_time);
  g_info("offset time(second): %d", controller->offset_time);

  g_info("temperature upper(degrees): %f", controller->temperature_upper);
  g_info("temperature lower(degrees): %f", controller->temperature_lower);
  g_info("humidity upper(percent): %f", controller->humidity_upper);
  g_info("humidity lower(percent): %f", controller->humidity_lower);
}

static GByteArray *
edispz_controller_create_timekeep_packet (EdispzController * controller, gchar *mac, gboolean permit)
{
  GByteArray *packet;
  GDateTime *date_time, *next_time;
  guint32 t, current_time, next_base_time;
  guchar *data;
  guchar header_template[] = {
    0x10, 0x82,                 /* Fixed */
    0x00, 0x00,                 /* TID */
    0x87,                       /* Command */
    0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,     /* Serial number of e-DISPZ */
    0x00,                       /* Reserved */
    0x25,                       /* Payload length */
  };
  guchar reserved = 0x00;

  g_return_val_if_fail (controller != NULL, NULL);

  packet = g_byte_array_new ();

  /* header */
  header_template[2] = controller->tid_up;
  header_template[3] = controller->tid_low;
  g_byte_array_append (packet, header_template, sizeof (header_template));
  if (!++controller->tid_low)
    controller->tid_up++;

  /* current time */
  date_time = g_date_time_new_now_local ();
  t = 0;
  t |= ((g_date_time_get_year (date_time) - 2000) << 26);
  t |= (g_date_time_get_month (date_time) << 22);
  t |= (g_date_time_get_day_of_month (date_time) << 17);
  t |= (g_date_time_get_hour (date_time) << 12);
  t |= (g_date_time_get_minute (date_time) << 6);
  t |= (g_date_time_get_second (date_time) << 0);
//if it has not own mac address, set time 0 to reject.
  if(permit != TRUE)
	  t = 0;
  current_time = t;
  current_time = GUINT_TO_BE (current_time);

  data = g_malloc (sizeof (guint32));
  memcpy (data, &current_time, 4);
  g_byte_array_append (packet, data, sizeof (guint32));
  g_free (data);

  /* next base time */
  next_time = g_date_time_add_seconds (date_time, (gdouble)controller->interval_time);
  t = 0;
  t |= ((g_date_time_get_year (next_time) - 2000) << 26);
  t |= (g_date_time_get_month (next_time) << 22);
  t |= (g_date_time_get_day_of_month (next_time) << 17);
  t |= (g_date_time_get_hour (next_time) << 12);
  t |= (g_date_time_get_minute (next_time) << 6);
  t |= (g_date_time_get_second (next_time) << 0);

  next_base_time = t;
  next_base_time = GUINT_TO_BE (next_base_time);

  g_date_time_unref(date_time);
  g_date_time_unref(next_time);

  data = g_malloc (sizeof (guint32));
  memcpy (data, &next_base_time, sizeof (guint32));
  g_byte_array_append (packet, data, sizeof (guint32));
  g_free (data);

  if (is_immediate_mode(controller->operation_mode)) {
    /* human senser wait time */
    g_byte_array_append (packet, (guint8 *) & controller->human_sensor_wait_time,
                         sizeof (guint));
  } else {
    /* interval */
    g_byte_array_append (packet, (guint8 *) & controller->interval_time,
                         sizeof (guint));
  }

  /* a request available time at the firmware update */
  t = 10;
  g_byte_array_append (packet, (guint8 *) & t, sizeof (guint32));

  if (is_immediate_mode(controller->operation_mode)) {
    /* heartbeat time */
    g_byte_array_append (packet, (guint8 *) & controller->heartbeat_time,
                         sizeof (gushort));
    /* door detect mode */
    g_byte_array_append (packet, (guint8 *) & controller->door_detect_mode,
                         sizeof (guchar));
    /* reserved */
    g_byte_array_append (packet, (guint8 *) & reserved,
                         sizeof (guchar));
  }

  {
    /* offset */
    Bp35a1Client *bp35a1 = controller->bp35a1_client;
    gint index;
    guint offset;
    index = bp35a1_client_whitelist_get_index(bp35a1, mac);
    offset = controller->offset_time * index;
    g_byte_array_append (packet, (guint8 *)&offset, sizeof (guint));
  }

  /* PSK of PANA */
  g_byte_array_append (packet, (guint8 *) controller->pana_psk,
      sizeof (controller->pana_psk));

  /* Mode select */
  g_byte_array_append (packet, (guint8 *) & controller->operation_mode,
      sizeof (controller->operation_mode));

  return packet;
}

static gboolean
edispz_controller_is_packet (GByteArray * packet)
{
  EdispzHeader *header = (EdispzHeader *)packet->data;

  g_return_val_if_fail (packet != NULL, FALSE);

  if (packet->len < EDISPZ_HEADER_SIZE)
    return FALSE;

  if (header->first.fixed[0] != 0x10 || header->first.fixed[1] != 0x82)
    return FALSE;

  if (packet->len < header->second.payload_len + EDISPZ_HEADER_SIZE)
    return FALSE;

  return TRUE;
}

static gboolean
edispz_controller_is_special_packet (GByteArray * packet)
{
  EdispzPacketType type;
  gboolean ret = FALSE;

  g_return_val_if_fail (packet != NULL, FALSE);

  type = edispz_packet_get_type (packet);
  if (type == EDISPZ_PACKET_SPECIAL)
    ret = TRUE;

  return ret;
}

static EdispzPacket *
new_edispz_packet(gchar *mac)
{
  EdispzPacket *packet;

  packet = g_slice_new0(EdispzPacket);
  packet->mac = g_strdup(mac);
  return packet;
}

void
delete_edispz_packet(EdispzPacket *packet)
{
  if(packet == NULL)
    return;

  g_byte_array_unref(packet->data);
  if (packet->mac)
    g_free(packet->mac);
  g_slice_free(EdispzPacket, packet);
}

static void
edispz_controller_loop (gpointer * controller)
{
  EdispzController *self;
  Bp35a1Client *bp35a1;
  GAsyncQueue *queue;

  Bp35a1Message **msg_array;
  Bp35a1Message *msg;
  gchar *mac;
  Bp35a1MessageType type;
  EdispzPacket *packet;
  gboolean permit_to_timekeep = TRUE;
  guint i;


  self = (EdispzController *) controller;
  queue = self->dest_queue;
  bp35a1 = self->bp35a1_client;

  g_mutex_lock (&self->lock);
  self->is_running = TRUE;
  g_mutex_unlock (&self->lock);

  print_config(self);

  while (self->is_running) {
    msg_array = bp35a1_client_receive_messages (bp35a1);

    if (msg_array != NULL) {
      for (i = 0; msg_array[i] != NULL; i++) {
        msg = msg_array[i];
        type = bp35a1_message_get_message_type (msg);
        switch (type) {
          case BP35A1_MESSAGE_ERXUDP:
	    mac  = bp35a1_message_get_mac_address(msg);
	    packet = new_edispz_packet(mac);
	    g_free(mac);
	    packet->data = bp35a1_client_get_message_data_permittable (bp35a1, msg, &permit_to_timekeep);
            if (packet->data) {
              if (edispz_controller_is_special_packet (packet->data)) {
                Bp35a1Message *timekeep_msg;
                GByteArray *timekeep_packet;
                timekeep_msg = bp35a1_message_new ();

                timekeep_packet = edispz_controller_create_timekeep_packet (self, packet->mac, permit_to_timekeep);
                timekeep_msg =
                 bp35a1_client_new_response_message (self->bp35a1_client, msg,
                                                     timekeep_packet);

                bp35a1_client_post_message (bp35a1, timekeep_msg);

                g_byte_array_free (timekeep_packet, TRUE);
              }
              if (!permit_to_timekeep){
                delete_edispz_packet(packet);
                continue;
              }
              g_async_queue_push (queue, packet);
            }
	    else
	      delete_edispz_packet(packet);
            break;
          default:
            break;
        }
        g_object_unref (msg_array[i]);
      }
      g_free (msg_array);
    }
  }

  g_thread_exit (NULL);
}

EdispzController *
edispz_controller_new (GAsyncQueue * dest_queue)
{
  guint i;
  EdispzController *ctrl;

  g_return_val_if_fail (dest_queue != NULL, NULL);

  ctrl = g_slice_new (EdispzController);

  ctrl->dest_queue = dest_queue;
  g_async_queue_ref (ctrl->dest_queue);
  ctrl->bp35a1_client = bp35a1_client_new ();
  ctrl->interval_time = 60;
  ctrl->offset_time = 1;
  ctrl->human_sensor_wait_time = 3;
  ctrl->heartbeat_time = 1;
  ctrl->door_detect_mode = 0;
  ctrl->operation_mode = 32;
  ctrl->temperature_upper = 100.00;
  ctrl->temperature_lower = -273.15;
  ctrl->humidity_upper = 100.00;
  ctrl->humidity_lower = 0.00;

  /* Please set PSK if use security authentication (PANA) */
  for (i = 0; i < 16; i++)
    ctrl->pana_psk[i] = 0x00;

  g_mutex_init (&ctrl->lock);
  ctrl->is_running = FALSE;

  return ctrl;
}

gint
edispz_controller_start (EdispzController * controller)
{
  gint ret;

  g_return_val_if_fail (controller != NULL, -1);

  ret = bp35a1_client_start (controller->bp35a1_client);
  if (ret < 0)
    return ret;

  bp35a1_client_debug_dump_registers (controller->bp35a1_client);

  controller->loop = g_thread_new ("e-DISP Z controller loop",
      (GThreadFunc) edispz_controller_loop, controller);

  return 0;
}

gint
edispz_controller_stop (EdispzController * controller)
{
  g_return_val_if_fail (controller != NULL, -1);

  g_mutex_lock (&controller->lock);
  controller->is_running = FALSE;
  g_mutex_unlock (&controller->lock);

  g_thread_join (controller->loop);

  bp35a1_client_stop (controller->bp35a1_client);

  return 0;
}

void
edispz_controller_free (EdispzController * controller)
{
  g_return_if_fail (controller != NULL);

  controller->is_running = FALSE;
  g_mutex_clear (&controller->lock);

  g_async_queue_unref (controller->dest_queue);
  g_free (controller->device);

  g_object_unref (controller->bp35a1_client);
}

EdispzPacketType
edispz_packet_get_type (GByteArray * packet)
{
  EdispzHeader *header = (EdispzHeader *)packet->data;
  EdispzPacketType packet_type;

  g_return_val_if_fail (packet != NULL, EDISPZ_PACKET_BROKEN);

  if (!edispz_controller_is_packet (packet))
    return EDISPZ_PACKET_BROKEN;

  packet_type = header->second.command & 0x0F;
  switch (packet_type) {
    case EDISPZ_PACKET_ACK_NACK:
      if (header->second.payload_len < EDISPZ_PACKET_ACK_NACK_LENGTH)
        packet_type = EDISPZ_PACKET_BROKEN;
      break;
    case EDISPZ_PACKET_TIMEKEEPING:
      if (header->second.payload_len < EDISPZ_PACKET_TIMEKEEPING_LENGTH)
        packet_type = EDISPZ_PACKET_BROKEN;
      break;
    case EDISPZ_PACKET_SPECIAL:
      if (header->second.payload_len < EDISPZ_PACKET_SPECIAL_LENGTH)
        packet_type = EDISPZ_PACKET_BROKEN;
      break;
    case EDISPZ_PACKET_TEST:
      if (header->second.payload_len < EDISPZ_PACKET_TEST_LENGTH)
        packet_type = EDISPZ_PACKET_BROKEN;
      break;
    case EDISPZ_PACKET_NORMAL:
      if (header->second.payload_len < EDISPZ_PACKET_NORMAL_LENGTH)
        packet_type = EDISPZ_PACKET_BROKEN;
      break;
    default:
        packet_type = EDISPZ_PACKET_UNKNOWN;
      break;
  }

  return packet_type;
}

static gchar *
edispz_ack_nack_packet_get_data (GByteArray * packet,
    EdispzPacketDataType data_type)
{
  switch (data_type) {
    case EDISPZ_PACKET_DATA_ACK_NACK:
    {
      guint8 *data;
      data = packet->data + EDISPZ_HEADER_SIZE;

      if (data[0] == 0xFF)
        return g_strdup ("ACK");
      else if (data[0] == 0x00)
        return g_strdup ("NACK");
    }
      break;
    default:
      break;
  }

  return NULL;
}

static gchar *
edispz_normal_packet_get_data (guint8 * data, EdispzPacketDataType data_type)
{
  switch (data_type) {
    case EDISPZ_PACKET_DATA_DOOR_IS_OPEN:
    {
      guint8 state;

      state = data[1] & 0x08;
      if (state == 0x08)
        return g_strdup_printf ("open");
      else if (state == 0x00)
        return g_strdup_printf ("close");
    }
    case EDISPZ_PACKET_DATA_DOOR_TIME:
    {
      guint16 time = 0;

      time |= (data[1] & 0x07) << 8;
      time |= data[0];
      return g_strdup_printf ("%02u:%02u", time / 60, time % 60);
    }
      break;
    case EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT:
    {
      return g_strdup_printf ("%u", data[2]);
    }
      break;
    case EDISPZ_PACKET_DATA_TEMPERATURE:
    {
      guint16 temp = 0;

      temp |= (data[3] & 0x3F) << 8;
      temp |= data[4];
      if (data[3] & 0x80 || temp == 0x3FFF){
	/* it means temparature field is unused.*/
	/* the absolute zero(=-273.15) - 1 as error value.*/
	return g_strdup_printf ("-274");
      }
      temp = temp << 2;
      return g_strdup_printf ("%f",
          (gfloat) (-46.85 + 175.72 * temp / 65536.0));
    }
      break;
    case EDISPZ_PACKET_DATA_HUMIDITY:
    {
      guint16 temp = 0;

      temp |= (data[5] & 0x0F) << 8;
      temp |= data[6];
      if (data[5] & 0x80 || temp == 0xFFF){
	/* it means humidity field is unused.*/
	/* -1 as error value.*/
	return g_strdup_printf ("-1");
      }
      temp = temp << 4;
      return g_strdup_printf ("%f", (gfloat) (-6.0 + 125.0 * temp / 65536.0));
    }
      break;
    case EDISPZ_PACKET_DATA_HUMAN_IS_REACT:
    {
      guint8 state;

      state = data[8] & 0x80;
      if (state == 0x80)
        return g_strdup_printf ("true");
      else if (state == 0x00)
        return g_strdup_printf ("false");
    }
      break;
    case EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT:
    {
      guint16 count = 0;

      count |= (data[8] & 0x0F) << 8;
      count |= data[7];
      return g_strdup_printf ("%u", count);
    }
      break;
    default:
      break;
  }

  return NULL;
}

static gchar *
edispz_special_packet_get_data (guint8 * data, EdispzPacketDataType data_type)
{
  switch (data_type) {
    case EDISPZ_PACKET_DATA_BATTERY:
    {
      return g_strdup_printf ("%u.%02u", data[0], data[1]);
    }
      break;
    case EDISPZ_PACKET_DATA_DOOR_IS_OPEN:
    case EDISPZ_PACKET_DATA_DOOR_TIME:
    case EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT:
    case EDISPZ_PACKET_DATA_TEMPERATURE:
    case EDISPZ_PACKET_DATA_HUMIDITY:
    case EDISPZ_PACKET_DATA_HUMAN_IS_REACT:
    case EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT:
    {
      return edispz_normal_packet_get_data (data + 2, data_type);
    }
      break;
    case EDISPZ_PACKET_DATA_FIRMWARE_VERSION:
    {
      return g_strdup_printf ("%c%c%c%c",
          data[11], data[12], data[13], data[14]);
    }
      break;
    case EDISPZ_PACKET_DATA_TIMEKEEPING_FLAG:
    {
      return g_strdup_printf ("%d", data[15]);
    }
      break;
    default:
      break;
  }

  return NULL;
}

gchar *
edispz_packet_get_data (GByteArray * packet, EdispzPacketDataType data_type)
{
  gchar *ret;
  guint8 *payload;

  g_return_val_if_fail (packet != NULL, NULL);

  payload = packet->data + EDISPZ_HEADER_SIZE;

  switch (edispz_packet_get_type (packet)) {
    case EDISPZ_PACKET_ACK_NACK:
      ret = edispz_ack_nack_packet_get_data (packet, data_type);
      break;
    case EDISPZ_PACKET_SPECIAL:
    case EDISPZ_PACKET_TEST:
      ret = edispz_special_packet_get_data (payload, data_type);
      break;
    case EDISPZ_PACKET_NORMAL:
      ret = edispz_normal_packet_get_data (payload, data_type);
      break;
    default:
      ret = NULL;
      break;
  }

  return ret;
}
