#include "common.h"
#include "fluentd-client.h"

G_DEFINE_TYPE (FluentdClient, fluentd_client, G_TYPE_OBJECT)

static void
fluentd_client_get_property (GObject    * object,
                             guint        property_id,
                             GValue     * value,
                             GParamSpec * pspec)
{
  switch (property_id)
    {
    case FLUENTD_CLIENT_PROP_SOCKNAME:
      g_value_set_string (value, fluentd_client_get_sockname ((FluentdClient *)object));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
fluentd_client_set_property (GObject      * object,
                             guint          property_id,
                             const GValue * value,
                             GParamSpec   * pspec)
{
  switch (property_id)
    {
    case FLUENTD_CLIENT_PROP_SOCKNAME:
      fluentd_client_set_sockname ((FluentdClient *)object, g_value_get_string (value));
      break;
    default:
      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, property_id, pspec);
      break;
    }
}

static void
fluentd_client_finalize (GObject * object)
{
  FluentdClient *self = (FluentdClient *)object;

  if (self->fd >= 0)
    close (self->fd);

  if (self->sockname)
    g_free (self->sockname);

  if (G_OBJECT_CLASS (fluentd_client_parent_class))
    G_OBJECT_CLASS (fluentd_client_parent_class)->finalize (object);
}

static void
fluentd_client_class_init (FluentdClientClass * klass)
{
  GObjectClass *gobject_class = G_OBJECT_CLASS (klass);

  gobject_class->get_property = fluentd_client_get_property;
  gobject_class->set_property = fluentd_client_set_property;
  gobject_class->finalize = fluentd_client_finalize;

  g_object_class_install_property (gobject_class, FLUENTD_CLIENT_PROP_SOCKNAME,
      g_param_spec_string ("sockname", "Socketname", "Soecket name",
          FLUENTD_CLIENT_SOCKNAME, G_PARAM_CONSTRUCT | G_PARAM_READWRITE | G_PARAM_STATIC_STRINGS));
}

static void
fluentd_client_init (FluentdClient * self)
{
  self->fd = -1;
}

FluentdClient *
fluentd_client_new (gchar * sockname)
{
  FluentdClient *ret;
  gchar *name = NULL;

  ret = g_object_new (FLUENTD_TYPE_CLIENT, NULL);

  if (sockname != NULL) {
    g_object_set ((GObject *) ret, "sockname", sockname, NULL);
  }

  if ((ret->fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0) {
    g_print ("socket\n");
    goto fail;
  }

  g_object_get ((GObject *) ret, "sockname", &name, NULL);

  bzero((char *)&ret->addr, sizeof(ret->addr));
  ret->addr.sun_family = AF_UNIX;
  strcpy(ret->addr.sun_path, name);

  if (connect(ret->fd, (struct sockaddr *)&ret->addr,
              sizeof(ret->addr.sun_family) + strlen(name)) < 0){
    g_print ("connect\n");
    goto fail;
  }

  g_free (name);

  return ret;

fail:
  if (name != NULL)
    g_free (name);

  g_object_unref (ret);
  return NULL;
}

gchar *
fluentd_client_get_sockname (FluentdClient * client)
{
  gchar *result;

  g_return_val_if_fail (FLUENTD_IS_CLIENT (client), NULL);

  result = client->sockname;

  return result;
}

void
fluentd_client_set_sockname (FluentdClient * client, const gchar * string)
{
  g_return_if_fail (FLUENTD_IS_CLIENT (client));
  g_return_if_fail (string != NULL);

  if (client->sockname != NULL) {
    if (strcmp(client->sockname, string) == 0)
      return;

    g_free (client->sockname);
  }

  client->sockname = g_strdup (string);
}

void
fluentd_client_post_json (FluentdClient * client, gchar * tag, gchar * json)
{
  const gchar format[] = "[ \"%s\", %d, %s ]";
  gchar *buf;

  g_return_if_fail (FLUENTD_IS_CLIENT (client));
  g_return_if_fail (json != NULL);

  if (tag == NULL)
    buf = g_strdup_printf (format, "sensor.value", 0, json);
  else
    buf = g_strdup_printf (format, tag, 0, json);

  xwrite (client->fd, buf, strlen(buf));
  g_free (buf);
}

gboolean
fluentd_client_connect (FluentdClient * client)
{
  gchar *name;

  g_return_val_if_fail (FLUENTD_IS_CLIENT (client), FALSE);

  name = (gchar *) g_object_get_data ((GObject *) client, "sockname");
  if (connect(client->fd, (struct sockaddr *)&client->addr,
              sizeof(client->addr.sun_family) + strlen(name)) < 0){
    return FALSE;
  }

  return TRUE;
}
