#include "edispz-controller.h"
#include "edispz-config.h"
#include "fluentd-client.h"
#include <getopt.h>

struct option_args {
  const char *config_file;
  const char *connection_list_file;
};

static struct option_args args;

static gboolean
load_config(EdispzController *controller, const gchar *filename)
{
  Bp35a1Client *client = controller->bp35a1_client;
  GKeyFile *keyfile;
  gchar *path = NULL;
  gchar *conf = NULL;
  gboolean ret = TRUE;

  keyfile = g_key_file_new ();
  if (filename == NULL)
    path = g_build_path("/", CONFIG_PATH, CONFIG_FILE, NULL);
  else
    path = g_build_path("/", filename, NULL);
    
  if (!g_key_file_load_from_file (keyfile, path, G_KEY_FILE_NONE, NULL)) {
    g_error("load config error %s", path);
    ret = FALSE;
    goto err;
  }
  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_TTY, NULL, NULL);
  if (conf == NULL) {
    g_error("\"tty=\" not found in %s", path);
    ret = FALSE;
    goto err;
  }
  bp35a1_client_set_device_name(client, conf);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_MODE, NULL, NULL);
  if (conf == NULL)
    g_warning("\"mode=\" not found in %s", path);
  else
    controller->operation_mode = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_INTERVAL_TIME, NULL, NULL);
  if (conf != NULL)
    controller->interval_time = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_OFFSET_TIME, NULL, NULL);
  if (conf != NULL)
    controller->offset_time = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_HUMAN_SENSOR_WAIT_TIME, NULL, NULL);
  if (conf != NULL)
    controller->human_sensor_wait_time = strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_HEART_BEAT_TIME, NULL, NULL);
  if (conf != NULL)
    controller->heartbeat_time = (gushort)strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_DOOR_DETECT_MODE, NULL, NULL);
  if (conf != NULL)
    controller->door_detect_mode = (guchar)strtoul(conf, NULL, 10);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_TEMPERATURE_UPPER, NULL, NULL);
  if (conf != NULL)
    controller->temperature_upper = (gfloat)g_ascii_strtod(conf, NULL);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_TEMPERATURE_LOWER, NULL, NULL);
  if (conf != NULL)
    controller->temperature_lower = (gfloat)g_ascii_strtod(conf, NULL);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_HUMIDITY_UPPER, NULL, NULL);
  if (conf != NULL)
    controller->humidity_upper = (gfloat)g_ascii_strtod(conf, NULL);

  conf = g_key_file_get_locale_string(keyfile, CONFG_CORE, CONFK_HUMIDITY_LOWER, NULL, NULL);
  if (conf != NULL)
    controller->humidity_lower = (gfloat)g_ascii_strtod(conf, NULL);

 err:
  g_key_file_free(keyfile);
  g_free(path);

  return ret;
}

static gint
load_connection(EdispzController *controller, const gchar *filename)
{
  Bp35a1Client *client = controller->bp35a1_client;
  gint ret;
  guint i;
  int err = 0;
  gchar *data;
  gchar **mac;
  struct stat st;
  gchar *path = NULL;

  if (filename == NULL)
    path = g_build_path("/", CONFIG_PATH, CONNECTION_FILE, NULL);
  else
    path = g_build_path("/", filename, NULL);

  g_debug("connection.list = %s",path);

  err = stat(path, &st);
  if (err){
    g_error ("load connection: Failed to stat '%s': %d", path, err);
    goto err;
  }

  if (g_file_get_contents(path, &data, NULL, NULL)) {
    mac = g_strsplit (data, "\n", -1);
    for (i = 0; mac[i]; i++) {
      if( !g_str_has_prefix(mac[i], "#")) {
	ret = bp35a1_client_whitelist_add (client, mac[i]);
	if (ret == 0)
	  g_debug ("whitelist += %s", mac[i]);
      }
    }
    g_strfreev(mac);
    g_free(data);
  }

 err:
  g_free (path);
  return err;
}

void dump_raw_data(GByteArray *data)
{
  guint i;

  for(i=0; i<data->len; i++)
    {
      g_print("%x", data->data[i]);
    }
  g_print("\n");
}

gchar *
build_event(gchar *events, gchar *event_name, gchar *value)
{
  static gchar *fmt = "{\"event_id\":\"%s\", \"value\":\"%s\" }";
  gchar *event, *composite;

  event = g_strdup_printf(fmt, event_name, value);
  if (events) {
    composite = g_strjoin(",", event, events, NULL);
    g_free(events);
    g_free(event);
  } else {
    composite = event;
  }

  return composite;
}

gchar *
build_json(gchar *id, gchar *events)
{
  static gchar *fmt = "{\"device_id\":\"%s\", \"data\":[%s]}";
  return g_strdup_printf(fmt, id, events);
}

static gchar *
edispz_decode_data(EdispzController *controller, EdispzPacketType type,
		   EdispzPacket *packet)
{
  gchar *json;
  gchar *mac_head, *mac_tail, *mac;
  gchar *value;
  gchar *events = NULL;

  mac_head = g_strndup(packet->mac, 6);
  mac_tail = g_strndup((packet->mac) + 10, 6);
  mac = g_strconcat(mac_head, mac_tail, NULL);
  g_free(mac_head);
  g_free(mac_tail);

  value = edispz_packet_get_data (packet->data,
				  EDISPZ_PACKET_DATA_DOOR_IS_OPEN);
  if (value) {
    events = build_event(events, "get_door", value);
    g_free(value);
  }

  value = edispz_packet_get_data (packet->data,
				  EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT);
  if (value) {
    events = build_event(events, "get_door_count_before", value);
    g_free(value);
  }

  value = edispz_packet_get_data (packet->data,
				  EDISPZ_PACKET_DATA_DOOR_TIME);
  if (value) {
    events = build_event(events, "get_door_time", value);
    g_free(value);
  }

  value = edispz_packet_get_data (packet->data,
				  EDISPZ_PACKET_DATA_TEMPERATURE);
  if (value) {
    gfloat v = (gfloat)g_ascii_strtod(value, NULL);
    if ((controller->temperature_lower <= v &&
	 controller->temperature_upper >= v))
      events = build_event(events, "get_temperature", value);
    g_free(value);
  }

  value = edispz_packet_get_data (packet->data,
				  EDISPZ_PACKET_DATA_HUMIDITY);
  if (value) {
    gfloat v = (gfloat)g_ascii_strtod(value, NULL);
    if ((controller->humidity_lower <= v &&
	 controller->humidity_upper >= v))
      events = build_event(events, "get_humidity", value);
    g_free(value);
  }

  value = edispz_packet_get_data (packet->data,
				  EDISPZ_PACKET_DATA_HUMAN_IS_REACT);
  if (value) {
    events = build_event(events, "get_human_react", value);
    g_free(value);
  }

  value = edispz_packet_get_data (packet->data,
				  EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT);
  if (value) {
    events = build_event(events, "get_human_react_count", value);
    g_free(value);
  }

  if (type == EDISPZ_PACKET_SPECIAL ||
      type == EDISPZ_PACKET_TEST) {
    value = edispz_packet_get_data (packet->data,
				    EDISPZ_PACKET_DATA_BATTERY);
    if (value) {
      events = build_event(events, "get_battery", value);
      g_free(value);
    }
  }

  json = build_json(mac, events);
  g_free(events);
  g_free(mac);

  return json;
}

void
fluentd_send_data(gchar *json)
{
  FluentdClient *c;

  c = fluentd_client_new (NULL);
  if (c == NULL) {
    g_warning("fail to connect fluentd server");
    return;
  }

  fluentd_client_post_json (c, NULL, json);
  g_object_unref (c);
  return;
}

void dump_data(GByteArray *data)
{
  EdispzPacketType type;

  type = edispz_packet_get_type (data);
  switch (type) {
  case EDISPZ_PACKET_ACK_NACK:
    g_print ("  EDISPZ_PACKET_DATA_ACK_NACK: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_ACK_NACK));
    break;
  case EDISPZ_PACKET_NORMAL:
    g_print ("  EDISPZ_PACKET_DATA_DOOR_IS_OPEN: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_IS_OPEN));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_TIME: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_TIME));
    g_print ("  EDISPZ_PACKET_DATA_TEMPERATURE: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_TEMPERATURE));
    g_print ("  EDISPZ_PACKET_DATA_HUMIDITY: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMIDITY));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_IS_REACT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_IS_REACT));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT));
    break;
  case EDISPZ_PACKET_SPECIAL:
  case EDISPZ_PACKET_TEST:
    g_print ("  EDISPZ_PACKET_DATA_BATTERY: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_BATTERY));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_IS_OPEN: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_IS_OPEN));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_OPEN_COUNT));
    g_print ("  EDISPZ_PACKET_DATA_DOOR_TIME: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_DOOR_TIME));
    g_print ("  EDISPZ_PACKET_DATA_TEMPERATURE: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_TEMPERATURE));
    g_print ("  EDISPZ_PACKET_DATA_HUMIDITY: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMIDITY));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_IS_REACT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_IS_REACT));
    g_print ("  EDISPZ_PACKET_DATA_HUMAN_COUNT: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_HUMAN_REACT_COUNT));
    g_print ("  EDISPZ_PACKET_DATA_FIRMWARE_VERSION: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_FIRMWARE_VERSION));
    g_print ("  EDISPZ_PACKET_DATA_TIMEKEEPING_FLAG: %s\n",
	     edispz_packet_get_data (data, EDISPZ_PACKET_DATA_TIMEKEEPING_FLAG));
    break;
  default:
    g_print ("unknown\n");
    break;
  }
}

static void usage(void)
{
  g_printerr("edispz-device-app [-c CONFIG FILE PATH] [-p CONNECTION LIST PATH]\n");
  g_printerr("  -c, --config=CONFIG FILE PATH\n");
  g_printerr("      default: /etc/edispz-device-app/edispz.conf\n");
  g_printerr("  -l, --list=CONNECTION LIST PATH\n");
  g_printerr("      default: /etc/edispz-device-app/connection.list\n");
  g_printerr("  -h, --help\n");
  g_printerr("      Print usage(this message) and exit.\n");
}

static int parse_args(int argc, char *argv[]/*, char *config_file, char *connection_list_file*/)
{
  struct option longopts[] = {
    {"conifg",required_argument, NULL, 'c' },
    {"list",required_argument, NULL, 'l' },
    {"help",no_argument , NULL, 'h' },
    {0,0,0,0},
  };
  int c;

  args.config_file = NULL;
  args.connection_list_file = NULL;

  while ((c = getopt_long(argc, argv, "c:l:h", longopts, NULL)) != EOF) {
    switch (c) {
    case 'c':
      args.config_file = optarg;
      break;
    case 'l':
      args.connection_list_file = optarg;
      break;
    case 'h':
      g_printerr("h\n");
      usage();
      return 1;
    default:
      g_printerr("error\n");
      return -1;
    }
  }

  return 0;
}

int
main(int argc, char *argv[])
{
  gint ret;
  GAsyncQueue *aqueue;
  EdispzController *edispz;
  EdispzPacket *packet;
  gchar *json;

  ret = parse_args(argc, argv);
  if (ret < 0) {
    usage();
    return -1;
  }
  if (ret) {
    return 0;
  }

  aqueue = g_async_queue_new ();
  edispz = edispz_controller_new (aqueue);
  if (edispz == NULL) {
    g_print ("controller new failed.\n");
  }

  ret = load_config(edispz, args.config_file);
  if (ret != TRUE) {
    g_error("no config file");
    return -1;
  }

  ret = edispz_controller_start (edispz);
  if (ret < 0) {
    g_print ("controller start failed.\n");
  }

  ret = load_connection (edispz, args.connection_list_file);
  if (ret != 0) {
    g_error("no e-DISPZ in connection list");
    return -2;
  }

  while (TRUE) {
    packet = (EdispzPacket *) g_async_queue_pop (aqueue);
    if (packet != NULL) {
      EdispzPacketType type = edispz_packet_get_type (packet->data);
      switch (type) {
      case EDISPZ_PACKET_NORMAL:
      case EDISPZ_PACKET_SPECIAL:
      case EDISPZ_PACKET_TEST:
	json = edispz_decode_data(edispz, type, packet);
	fluentd_send_data(json);
	g_free(json);
	break;
      default:
	/* ignore packet */
	break;
      }
      delete_edispz_packet(packet);
    }
  }

  return 0;
}

